#-*- coding: utf-8

import sys

def get_NEs(ary):
    '''
    Input:
    ary: list of iob2 tags
    Output:
    ret: dict of tagtype -> set of (tagtype, start, end)
    '''
    ret = {}
    ary += ['O']
    tagtype = None
    for i, t in enumerate(ary):
        if t in ['X', '[PAD]']:
            continue
        if t.endswith('-I'):
            continue
        if tagtype is not None:
            if not tagtype in ret:
                ret[tagtype] = set()
            ret[tagtype].add((start, i-1))
            tagtype = None
        if t != 'O' and t != '[PAD]':
            if t.startswith('B-') or t.startswith('I-'):
                tagtype = t[2:]
            else:
                tagtype = t[:-2]
            start = i
    return ret

def tagwise_precision_recall(predss, goldss):

    ret = {}
    for preds, golds in zip(predss, goldss):
        tag_set = set(preds.keys())|set(golds.keys())
        # tagtype -> (true positive, false positive, false negative)
    
        for tagtype in tag_set:
            gold = golds[tagtype] if tagtype in golds else set()
            pred = preds[tagtype] if tagtype in preds else set()
            if not tagtype in ret:
                ret[tagtype] = [0 for _ in range(4)]
            ret[tagtype][0] += len(gold&pred)
            ret[tagtype][1] += len(pred-gold)
            ret[tagtype][2] += len(gold-pred)
            ret[tagtype][3] += len(gold)

    return ret

def calc_fscore(precision_recall, print_fullscore=True, print_tagscore=False):

    format_str = '{:7} {:>7} {:>7} {:>7}'
    format_float = '{:7} {:>7.5f} {:>7.5f} {:>7.5f}'
    if print_tagscore:
        print(format_str.format('tagtype', 'prec.', 'recl.', 'Fscore'))
    full_cnt = [0 for _ in range(3)]
    f1s = []
    for tagtype, counts in sorted(precision_recall.items(), key=lambda x:x[1][3], reverse=True):
        tp, fp, fn, cnt = counts
        prec = tp/max(tp+fp, 1)
        recl = tp/max(tp+fn, 1)
        fscr = 2*recl*prec/(recl+prec) if recl+prec>0 else 0
        if print_tagscore:
            print(format_float.format(tagtype, prec, recl, fscr) + ' N = {}'.format(cnt))
        full_cnt[0] += tp
        full_cnt[1] += fp
        full_cnt[2] += fn
        f1s.append(fscr)
    if print_tagscore:
        print('======================')
    ftp=full_cnt[0]
    ffp=full_cnt[1]
    fnp=full_cnt[2]
    fprec=ftp/max(ftp+ffp, 1)
    frecl=ftp/max(ftp+fnp, 1)
    ffscr=2*frecl*fprec/(frecl+fprec) if frecl+fprec>0 else 0
    if print_fullscore:
        print('micro F1 = {:.5f} (prec = {:.5f} recl = {:.5f})'.format(ffscr, fprec, frecl))
        print('macro F1 = {:.5f}'.format(sum(f1s)/len(f1s) if len(f1s) > 0 else 0.0))
        print('=====')
    return ffscr, sum(f1s)/len(f1s) if len(f1s) > 0 else 0.0

def F1score(golds, preds, print_fullscore=True, print_tagscore=False):
    pred_NEs = [get_NEs(pred) for pred in preds]
    gold_NEs = [get_NEs(gold) for gold in golds]
    pr = tagwise_precision_recall(pred_NEs, gold_NEs)
    score = calc_fscore(pr, print_fullscore=print_fullscore, print_tagscore=print_tagscore)
    sys.stdout.flush()
    return score

if __name__=='__main__':

    argvs = sys.argv
    argc = len(argvs)

    if argc != 3:
        sys.stderr.write('Usage: python accuracy.py <gold iob2 file> <predict iob2 file>\n')
        exit(1)

    gold_fname = argvs[1]
    pred_fname = argvs[2]

    with open(gold_fname, 'r') as gf, open(pred_fname, 'r') as pf:

        gold_tags_full = [[w.split('/')[1:] for w in gl.rstrip().split(' ')] for gl in gf]
        n_tags = max([max([len(t) for t in gt]) for gt in gold_tags_full])

        for i in range(n_tags):
            gold_tags = [[t[i] for t in gt] for gt in gold_tags_full]
            pred_tags = [[w.split('/')[i+1] for w in pl.rstrip().split(' ')] for pl in pf]

            F1score(gold_tags, pred_tags, print_fullscore=True, print_tagscore=True)
        

