#-*- coding: utf-8 -*-
# based on https://github.com/Louis-udm/NER-BERT-CRF

import os
import sys
import json
import pickle
import random
import time
import shutil
import math

import numpy as np
import torch
import torch.nn as nn
from transformers import AutoModel, AutoConfig, AutoTokenizer, PreTrainedTokenizerFast
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from tokenizers import Tokenizer, normalizers, pre_tokenizers
from tokenizers.models import WordLevel
from tokenizers.processors import TemplateProcessing
from tokenizers.trainers import WordLevelTrainer

from accuracy import F1score

torch.multiprocessing.set_sharing_strategy('file_system')

TO_NONBLOCKING=True
NUM_WORKERS=0 #os.cpu_count()

def to_device(data, device, **kwargs):

    if hasattr(data, 'to'):
        return data.to(device, **kwargs)
    if isinstance(data, list):
        return [to_device(d, device, **kwargs) for d in data]
    return None

def make_whitespace_tokenizer(corpus, vocab_size=32_000, do_strip_accents=False, save_filename='tokenizer.json'):

    tokenizer = Tokenizer(WordLevel(unk_token="[UNK]"))
    tokenizer_normalizer_task_list = [normalizers.NFD(), normalizers.Lowercase()]
    if do_strip_accents:
        tokenizer_normalizer_task_list.append(normalizers.StripAccents())
    tokenizer.normalizer = normalizers.Sequence(tokenizer_normalizer_task_list)
    tokenizer.pre_tokenizer = pre_tokenizers.Sequence([pre_tokenizers.Whitespace(), pre_tokenizers.Digits(individual_digits=True)])

    trainer = WordLevelTrainer(vocab_size=vocab_size, special_tokens=["[UNK]", "[CLS]", "[SEP]", "[PAD]", "[MASK]"])
    tokenizer.train_from_iterator(corpus, trainer=trainer)
    tokenizer.save(save_filename)

class ModelOption():
    def __init__(self, option):

        # model definition
        self.bert_dir = option['bert_dir'] if 'bert_dir' in option else 'bert-base-uncased'
        self.bert_config = option['bert_config'] if 'bert_config' in option else None
        self.encoder_type = option['encoder_type'] if 'encoder_type' in option else 'bert'
        self.use_dict = option['use_dict'] if 'use_dict' in option else False
        # training hyper-parameter
        self.batch_size = option['batch_size'] if 'batch_size' in option else 32
        self.lr = option['lr'] if 'lr' in option else 2e-5
        self.num_epoch = option['num_epoch'] if 'num_epoch' in option else 20
        self.num_iteration = option['num_iteration'] if 'num_iteration' in option else None
        self.use_scheduler = option['use_scheduler'] if 'use_scheduler' in option else False
        self.lambda_pointwise = option['lambda_pointwise'] if 'lambda_pointwise' in option else 0.5
        ## for whitespace tokenizer
        self.use_whitespace_split = option['use_whitespace_split'] if 'use_whitespace_split' in option else False
        self.vocab_size = option['vocab_size'] if 'vocab_size' in option else 32_000
        self.do_strip_accents = option['do_strip_accents'] if 'do_strip_accents' in option else False
        self.whitespace_json_path = None
        # evaluation hyper-parameter
        self.eval_batch_size = option['eval_batch_size'] if 'eval_batch_size' in option else 64
        # utils
        self.quiet_train = option['quiet_train'] if 'quiet_train' in option else True
        self.quiet_valid = option['quiet_valid'] if 'quiet_valid' in option else True
        self.quiet_test = option['quiet_test'] if 'quiet_test' in option else True
        # training mode
        self.do_pre = option['do_pre'] if 'do_pre' in option else False
        self.do_onlydict = option['do_onlydict'] if 'do_onlydict' in option else False
        self.do_crfonly = option['do_crfonly'] if 'do_crfonly' in option else False
        self.target_tag = option['target_tag'] if 'target_tag' in option else None

    def update(self, opts):
        if 'bert_dir' in opts: self.bert_dir = opts['bert_dir']
        if 'bert_config' in opts: self.bert_config = opts['bert_config']
        if 'whitespace_json_path' in opts: self.whitespace_json_path = opts['whitespace_json_path']

    def delete_file_path(self):

        self.bert_dir = None
        self.bert_config = None
        self.whitespace_json_path = None

def label_correction(labels):

    '''
    Remove special tokens
    Change internal tags without begin tags
    '''
    label_previous = ['X']
    label_other = ['[SEP]', '[CLS]', '[UNK]', '[PAD]']
    
    for i in range(len(labels)):
        cur_label = labels[i]
        if cur_label in label_previous:
            if i > 0:
                cur_label = labels[i-1]
                if cur_label.startswith('B-'):
                    cur_label = 'I'+cur_label[1:]
                if cur_label.endswith('-B'):
                    cur_label = cur_label[:-1]+'I'
            else:
                cur_label = 'O'
        if cur_label in label_other:
            cur_label = 'O'
        if cur_label.startswith('I-'):
            if i==0 or cur_label[1:]!=labels[i-1][1:]:
                cur_label = 'B'+cur_label[1:]
        if cur_label.endswith('-I'):
            if i==0 or cur_label[:-1]!=labels[i-1][:-1]:
                cur_label = cur_label[:-1]+'B'
        labels[i] = cur_label

    return labels
            

def log_sum_exp_batch(log_Tensor, axis=-1):
    return torch.max(log_Tensor, axis)[0] + torch.log(torch.exp(log_Tensor - torch.max(log_Tensor, axis)[0].view(log_Tensor.shape[0], -1, 1)).sum(axis))

class InputExample(object):

    def __init__(self, words, labels):
        '''
        data class
        Args:
          words: list[int] : list of word ids
          labels: list[int] : list of tag ids
        '''
        self.words = words
        self.labels = labels

class InputFeatures(object):

    def __init__(self, input_ids, input_mask, segment_ids, predict_mask, label_ids, in_dict, full_label, not_full_pad):
        '''
        feature class for InputExample
        Args:
          input_ids: list[int] : list of word ids
          input_mask: list[int]
          segment_ids: list[int]
          predict_mask: list[int]
          label_ids: list[int]
          in_dict: list[list[str]]
        '''
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.predict_mask = predict_mask
        self.label_ids = label_ids
        self.in_dict = in_dict
        self.full_label = full_label
        self.not_full_pad = not_full_pad

class CRFDataset(torch.utils.data.Dataset):

    def __init__(self, dataset, tokenizer, label_map, tag_index, dictionary={}, max_seq_length=None, predict_subword=False, iob_prefix=False):

        examples = [InputExample(w, l) for w, l in dataset]

        self.examples = examples
        self.tokenizer = tokenizer
        self.label_map = label_map
        self.dictionary = dictionary
        self.max_seq_length = max_seq_length
        self.predict_subword = predict_subword
        self.iob_prefix = iob_prefix
        self.tag_index = tag_index

    def __len__(self):

        return len(self.examples)

    def __getitem__(self, idx):

        feat = example2feature(self.examples[idx], self.tokenizer, self.label_map, self.dictionary, self.tag_index, self.max_seq_length, predict_subword=self.predict_subword, iob_prefix=self.iob_prefix)
        return feat.input_ids, feat.input_mask, feat.segment_ids, feat.predict_mask, feat.label_ids, feat.in_dict, feat.full_label, feat.not_full_pad

    @classmethod
    def pad(cls, batch):

        seqlen_list = [len(sample[0]) for sample in batch]
        maxlen = np.array(seqlen_list).max()

        tag_in_dict = [sum(sum([sample[5][i] for sample in batch], []), []) for i in range(len(batch[0][5]))]
        maxtag = [max(tag_in_dict[i])+1 if len(tag_in_dict[i]) > 0 else 1 for i in range(len(tag_in_dict))]

        f = lambda x, seqlen: [sample[x] + [0] * (seqlen - len(sample[x])) for sample in batch]
        f2 = lambda x: [sample[x] for sample in batch]
        f3 = lambda x, j, seqlen, maxtag: [[[i+1 if i in vi else 0 for vi in sample[x][j]] + [0] * (seqlen - len(sample[x][j])) for sample in batch] for i in range(maxtag[j])]
        f4 = lambda x, i, seqlen: [sample[x][i] + [0] * (seqlen - len(sample[x][i])) for sample in batch]

        input_ids_list = torch.LongTensor(f(0, maxlen))
        input_mask_list = torch.LongTensor(f(1, maxlen))
        segment_ids_list = torch.LongTensor(f(2, maxlen))
        predict_mask_list = torch.BoolTensor(f(3, maxlen))
        label_ids_list = [torch.LongTensor(f4(4, i, maxlen)) for i in range(len(maxtag))]
        in_dict_list = [torch.LongTensor(f3(5, j, maxlen, maxtag)) for j in range(len(maxtag))]
        full_label_list = torch.BoolTensor(f2(6))
        not_full_pad_list = torch.BoolTensor(f2(7))

        return input_ids_list, input_mask_list, segment_ids_list, predict_mask_list, label_ids_list, in_dict_list, full_label_list, not_full_pad_list

def example2feature(example, tokenizer, label_map, dictionary=None, tag_index=None, max_seq_length=None, predict_subword=False, iob_prefix=False):

    def check_dictionary_word(da, db):

        if len(da) != len(db):
            return False

        for wa, wb in zip(da, db):
            if wa.lower() != wb.lower():
                return False

        return True
    '''
        if [wi.lower() for wi in da] == [wi.lower() for wi in db]:
            return True

        return False
    '''
    
    add_label = 'X'
    tokens = ['[CLS]']
    predict_mask = [0]
    label_ids = [[label_map[i]['[CLS]']] for i in range(len(label_map))]
    in_dict = [[[] for _ in example.words] for i in range(len(label_map))]
    for i, w in enumerate(example.words):
        sub_words = tokenizer.tokenize(w)
        if not sub_words:
            sub_words = ['[UNK]']
        tokens.extend(sub_words)
        for k in range(len(dictionary)):
            if dictionary[k] is not None:
                for tagtype, wordlist in dictionary[k].items():
                    for dw in wordlist:
                        #if example.words[i:i+len(dw)] == dw:
                        if check_dictionary_word(example.words[i:i+len(dw)], dw):
                            for j in range(i, i+len(dw)):
                                in_dict_tag_idx = tag_index[k][tagtype]*2
                                if i==j:
                                    in_dict_tag_idx += 1
                                if tagtype in tag_index[k] and not in_dict_tag_idx in in_dict[k][j]:
                                    in_dict[k][j].append(in_dict_tag_idx)
                            break
        
        for j in range(len(sub_words)):
            if j == 0:
                predict_mask.append(1)
                for k in range(len(label_ids)):
                    if len(example.labels) <= k:
                        label_ids[k].append(0)
                    elif example.labels[k][i] in label_map[k]:
                        label_ids[k].append(max(0, label_map[k][example.labels[k][i]]))
                    else:
                        label_ids[k].append(label_map[k]['O'])
            else:
                if predict_subword:
                    predict_mask.append(0)
                    for k in range(len(label_ids)):
                        subword_tag = example.labels[k][i]
                        if subword_tag.startswith('B-'):
                            subword_tag = 'I' + subword_tag[1:]
                        if subword_tag.endswith('-B'):
                            subword_tag = subword_tag[:-1] + 'I'
                        label_ids[k].append(max(0, label_map[k][subword_tag]))
                else:
                    predict_mask.append(0)
                    for k in range(len(label_ids)):
                        if len(example.labels) <= k:
                            label_ids[k].append(0)
                        else:
                            label_ids[k].append(max(0, label_map[k][add_label]))

    if max_seq_length is not None and len(tokens) > max_seq_length - 1:
        #print('Example is too long, length is {}, truncated to {}!'.format(len(tokens), max_seq_length))
        tokens = tokens[0:(max_seq_length - 1)]
        predict_mask = predict_mask[0:(max_seq_length - 1)]
        label_ids = [label_ids[k][0:(max_seq_length - 1)] for k in range(len(label_ids))]
        in_dict = [in_dict[k][0:(max_seq_length - 1)] for k in range(len(in_dict))]

    tokens.append('[SEP]')
    predict_mask.append(0)
    for k in range(len(label_ids)):
        label_ids[k].append(label_map[k]['[SEP]'])

    input_ids = tokenizer.convert_tokens_to_ids(tokens)
    segment_ids = [0] * len(input_ids)
    input_mask = [1] * len(input_ids)

    full_label = [False if k >= len(example.labels) else not '[PAD]' in example.labels[k] for k in range(len(label_map))]
    not_full_pad = [False if k >= len(example.labels) else len([e for e in example.labels[k] if not e in ['[PAD]', '[CLS]', '[SEP]', 'X']]) > 0 for k in range(len(label_map))]

    feat = InputFeatures(
        input_ids=input_ids,
        input_mask=input_mask,
        segment_ids=segment_ids,
        predict_mask=predict_mask,
        label_ids=label_ids,
        in_dict=in_dict,
        full_label=full_label,
        not_full_pad=not_full_pad,
    )
    return feat

def is_iob(tag, iob_prefix):

    if iob_prefix:
        if tag.startswith('B-'):
            return tag[2:], 'B'
        if tag.startswith('I-'):
            return tag[2:], 'I'
        return tag, 'O'
    if tag.endswith('-B'):
        return tag[:-2], 'B'
    if tag.endswith('-I'):
        return tag[:-2], 'I'
    return tag, 'O'

def worker_init_fn(worker_id):
    random.seed(worker_id)

class PositionalEncoding(nn.Module):

    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        
        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))
        pe = torch.zeros(max_len, 1, d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    def forward(self, x):
        """
        Args:
            x: Tensor, shape [seq_len, batch_size, embedding_dim]
        """
        x = x + self.pe[:x.size(0)]
        return self.dropout(x)

class CRF(nn.Module):

    def __init__(self, encoder, start_label_id, stop_label_id, label_map, tokenizer, dropout_rate=0.25, dictionary = {}, use_dict = False, device='cpu'):
        super(CRF, self).__init__()
        self.encoder = encoder
        self.hidden_size = self.encoder.get_hidden_size()
        self.start_label_id = start_label_id
        self.stop_label_id = stop_label_id
        self.label_map = label_map
        self.label_list = [[v[0] for v in sorted(self.label_map[i].items(), key=lambda x:x[1]) if v[1] >= 0] for i in range(len(self.label_map))]
        self.dropout_rate = dropout_rate
        self.device = device
        self.iob_prefix = False
        self.use_dict = use_dict
        self.tokenizer = tokenizer
        self.dictionary = dictionary
        self.dict_embed = 32

        self.dropout = nn.Dropout(dropout_rate)
        self.hidden2label = nn.ModuleList([
            nn.Linear(in_features=self.hidden_size+self.dict_embed, out_features=len(self.label_list[i]), bias=True) for i in range(len(self.label_list))
        ])
        self.ILLEGAL_SCORE = -10000.

        self.transitions = [
            nn.Parameter(torch.randn(len(self.label_list[i]), len(self.label_list[i]))).to(self.device, non_blocking=TO_NONBLOCKING) for i in range(len(self.label_list))
        ]
        for i in range(len(self.label_list)):
            self.transitions[i].data[start_label_id[i], :] = self.ILLEGAL_SCORE
            self.transitions[i].data[:, stop_label_id[i]] = self.ILLEGAL_SCORE

        self.tag_index = [{tag: idx for idx, tag in enumerate(self.dictionary[i].keys())} for i in range(len(self.dictionary))]

        if self.use_dict:
            self.tag_embed = nn.ModuleList([
                nn.Embedding(len(self.tag_index[i])*2+1, embedding_dim=32, padding_idx=None) for i in range(len(self.tag_index))
            ])

            tag_transformer_encoder_layer = nn.TransformerEncoderLayer(
                d_model=self.dict_embed,
                nhead=8,
                dropout=self.dropout_rate
            )
            self.tag_transformer = nn.ModuleList([
                nn.TransformerEncoder(
                    encoder_layer=tag_transformer_encoder_layer,
                    num_layers=3,
                ) for _ in range(len(self.tag_index))
            ])
            self.pos_encoder = PositionalEncoding(self.dict_embed)
                    
        self.option = {}

        for h2l in self.hidden2label:
            nn.init.xavier_uniform_(h2l.weight)
            nn.init.constant_(h2l.bias, 0.)

        self.to(self.device, non_blocking=TO_NONBLOCKING)

    def _forward_alg(self, feats, tag_i):

        T = feats.shape[1]
        batch_size = feats.shape[0]

        log_alpha = torch.zeros((batch_size, 1, len(self.label_list[tag_i]))).fill_(self.ILLEGAL_SCORE).to(self.device, non_blocking=TO_NONBLOCKING)
        log_alpha[:, 0, self.start_label_id[tag_i]] = 0

        for t in range(1, T):
            log_alpha = (log_sum_exp_batch(self.transitions[tag_i] + log_alpha, axis=-1) + feats[:, t]).unsqueeze(1)
        log_prob_all_barX = log_sum_exp_batch(log_alpha)

        return log_prob_all_barX

    def _get_input_features(self, input_ids, segment_ids, input_mask, in_dict, tag_i, only_dict=False):

        if self.encoder._get_max_length() is not None and self.encoder._get_max_length() < input_ids.size(1):

            orig_input_ids = input_ids
            orig_segment_ids = segment_ids
            orig_input_mask = input_mask

            p_input_ids_list = []
            p_segment_ids_list = []
            p_input_mask_list = []

            max_encoder_len = self.encoder._get_max_length()
            part_l = [i for i in range(0, input_ids.size(1), max_encoder_len)]

            for b in range(input_ids.size(0)):

                for pi in part_l:
                    pj = min(pi+max_encoder_len, input_ids.size(1))
                    p_input_ids_list.append(torch.cat([input_ids.data[b,pi:pj], torch.LongTensor([0]*(max_encoder_len-(pj-pi))).to(self.device, non_blocking=TO_NONBLOCKING)]))
                    p_segment_ids_list.append(torch.cat([segment_ids.data[b,pi:pj], torch.LongTensor([0]*(max_encoder_len-(pj-pi))).to(self.device, non_blocking=TO_NONBLOCKING)]))
                    p_input_mask_list.append(torch.cat([input_mask.data[b,pi:pj], torch.LongTensor([0]*(max_encoder_len-(pj-pi))).to(self.device, non_blocking=TO_NONBLOCKING)]))

            input_ids = torch.stack(p_input_ids_list)
            segment_ids = torch.stack(p_segment_ids_list)
            input_mask = torch.stack(p_input_mask_list)
        
        # output from encoder (BERT)
        out_encoder = self.encoder._get_features(input_ids, segment_ids, input_mask)

        if 'orig_input_ids' in locals():

            out_encoder = torch.reshape(out_encoder, (orig_input_ids.size(0), -1, out_encoder.size(2)))[:,:orig_input_ids.size(1),:]

        maxlen = out_encoder.size(1)

        # embedding for NE in dictionary
        if self.use_dict:

            embed = sum([self.tag_embed[tag_i](in_dict[i])[:,:maxlen] for i in range(in_dict.size(0))]) * math.sqrt(self.dict_embed)
            embed = self.pos_encoder(embed)
            tag_out = self.tag_transformer[tag_i](embed)
            if only_dict:
                out_encoder = torch.cat((torch.zeros(out_encoder.size(), device=tag_out.device), tag_out), 2)
            else:
                out_encoder = torch.cat((out_encoder, tag_out), 2)

        else:
            out_encoder = torch.cat((out_encoder, torch.zeros((out_encoder.size(0), out_encoder.size(1), self.dict_embed), device=out_encoder.device)), 2)

        out_encoder = self.dropout(out_encoder)
        input_feats = self.hidden2label[tag_i](out_encoder)

        return input_feats

    def _score_sentence(self, feats, label_ids, tag_i):

        T = feats.shape[1]
        batch_size = feats.shape[0]

        batch_transitions = self.transitions[tag_i].expand(batch_size, len(self.label_list[tag_i]), len(self.label_list[tag_i]))
        batch_transitions = batch_transitions.flatten(1)

        score = torch.zeros((feats.shape[0], 1)).to(self.device, non_blocking=TO_NONBLOCKING)

        for t in range(1, T):
            score = score \
                + batch_transitions.gather(-1, (label_ids[:, t] * len(self.label_list[tag_i]) + label_ids[:, t-1]).view(-1, 1)) \
                + feats[:, t].gather(-1, label_ids[:, t].view(-1, 1)).view(-1, 1)

        return score

    def _viterbi_decode(self, feats, tag_i):

        T = feats.shape[1]
        batch_size = feats.shape[0]

        log_delta = torch.Tensor(batch_size, 1, len(self.label_list[tag_i])).fill_(self.ILLEGAL_SCORE).to(self.device, non_blocking=TO_NONBLOCKING)
        log_delta[:, 0, self.start_label_id[tag_i]] = 0

        psi = torch.zeros((batch_size, T, len(self.label_list[tag_i])), dtype=torch.long).to(self.device, non_blocking=TO_NONBLOCKING)
        for t in range(1, T):
            log_delta, psi[:, t] = torch.max(self.transitions[tag_i] + log_delta, -1)
            log_delta = (log_delta + feats[:, t]).unsqueeze(1)
        path = torch.zeros((batch_size, T), dtype=torch.long).to(self.device, non_blocking=TO_NONBLOCKING)
        max_logLL_allz_allx, path[:, -1] = torch.max(log_delta.squeeze(), -1)

        for t in range(T-2, -1, -1):
            path[:, t] = psi[:, t+1].gather(-1, path[:, t+1].view(-1, 1)).squeeze()

        return max_logLL_allz_allx, path

    def neg_log_likelihood(self, input_ids, segment_ids, input_mask, label_ids, in_dict, full_label, only_dict=False):

        loss = []
        for i, (lids, in_d) in enumerate(zip(label_ids, in_dict)):
            full_label_tag = torch.tensor([full_label[j][i] for j in range(full_label.size(0))])
            full_label_i = torch.nonzero(full_label_tag).view(-1)

            if sum(full_label_i) == 0:
                continue
            in_d_f = torch.stack([in_d[j][full_label_i] for j in range(in_d.size(0))])
            input_feats = self._get_input_features(input_ids[full_label_i], segment_ids[full_label_i], input_mask[full_label_i], in_d_f, i, only_dict=only_dict)
            forward_score = self._forward_alg(input_feats, i)
            gold_score = self._score_sentence(input_feats, lids[full_label_i], i)
            loss.append(torch.mean(forward_score - gold_score))
        return sum(loss)/len(loss) if len(loss) > 0 else 0

    def neg_log_likelihood_pointwise(self, input_ids, segment_ids, input_mask, label_ids, in_dict, not_full_pad, only_dict=False):

        loss = []
        pointwise_loss = nn.CrossEntropyLoss()
        for i, (lids, in_d) in enumerate(zip(label_ids, in_dict)):
            not_full_pad_tag = torch.tensor([not_full_pad[j][i] for j in range(not_full_pad.size(0))])
            not_full_pad_i = torch.nonzero(not_full_pad_tag).view(-1)
            if sum(not_full_pad_i) == 0:
                continue
            in_d_f = torch.stack([in_d[j][not_full_pad_i] for j in range(in_d.size(0))])
            input_feats = self._get_input_features(input_ids[not_full_pad_i], segment_ids[not_full_pad_i], input_mask[not_full_pad_i], in_d_f, i, only_dict=only_dict)
            loss.append(pointwise_loss(input_feats.view(-1, input_feats.size(2)), lids[not_full_pad_i].view(-1)))
            
        return sum(loss)/len(loss) if len(loss) > 0 else 0

    def forward(self, input_ids, segment_ids, input_mask, in_dict):

        ret_list = []
        for i, in_d in enumerate(in_dict):
            input_feats = self._get_input_features(input_ids, segment_ids, input_mask, in_d, i)
            score_label_seq_ids = self._viterbi_decode(input_feats, i)
            ret_list.append(score_label_seq_ids)

        return ret_list

    def predict(self, texts, batch_size=100):

        n_class = len(self.label_map)
        input_data = [
            (t.split(' '), [['O' for _ in t.split(' ')] for __ in range(n_class)]) for t in texts
        ]
        dataset = CRFDataset(input_data, self.tokenizer, self.label_map, self.tag_index, dictionary=self.dictionary, max_seq_length=None)
        dataloader = torch.utils.data.DataLoader(dataset=dataset, batch_size=batch_size, shuffle=False, num_workers=NUM_WORKERS, worker_init_fn=worker_init_fn, collate_fn=CRFDataset.pad)
        labels = []
        
        self.eval()
        with torch.no_grad():
            for batch in dataloader:
                batch = tuple(to_device(t, self.device, non_blocking=TO_NONBLOCKING) for t in batch)
                input_ids, input_mask, segment_ids, predict_mask, label_ids, in_dict, _, __ = batch

                ret_list = self.forward(input_ids, segment_ids, input_mask, in_dict)
                predict_label = [
                    [
                        label_correction(
                            [self.label_list[i][li.item()] for li, mask_i in zip(predicted, mask) if mask_i]
                        )
                        for predicted, mask in zip(ret[1], predict_mask)
                    ] for i, ret in enumerate(ret_list)
                ]
                predict_label_rev = [[predict_label[i][j] for i in range(len(predict_label))] for j in range(len(predict_label[0]))] if len(predict_label) > 0 else None

                if predict_label_rev is not None:
                    labels.extend(predict_label_rev)

        return labels

    def train_ner(self, datasets, label_set, ne_dict, option):

        # dataset
        train_dataset = CRFDataset(datasets['train'], self.tokenizer, self.label_map, self.tag_index, dictionary=ne_dict, max_seq_length=self.encoder._get_max_length())
        valid_dataset = CRFDataset(datasets['valid'], self.tokenizer, self.label_map, self.tag_index, dictionary=ne_dict, max_seq_length=None) if 'valid' in datasets and len(datasets['valid']) > 0 else None
        test_dataset = CRFDataset(datasets['test'], self.tokenizer, self.label_map, self.tag_index, dictionary=ne_dict, max_seq_length=None) if 'test' in datasets and len(datasets['test']) > 0 else None

        train_dataloader = torch.utils.data.DataLoader(dataset=train_dataset,
                                                      batch_size=option.batch_size,
                                                      shuffle=True,
                                                      num_workers=NUM_WORKERS,
                                                      pin_memory=True,
                                                      worker_init_fn=worker_init_fn,
                                                      collate_fn=CRFDataset.pad)
        valid_dataloader = torch.utils.data.DataLoader(dataset=valid_dataset,
                                                       batch_size=option.eval_batch_size,
                                                       shuffle=False,
                                                       num_workers=NUM_WORKERS,
                                                       pin_memory=True,
                                                       worker_init_fn=worker_init_fn,
                                                       collate_fn=CRFDataset.pad) if valid_dataset is not None else None
        test_dataloader = torch.utils.data.DataLoader(dataset=test_dataset,
                                                      batch_size=option.eval_batch_size,
                                                      shuffle=False,
                                                      num_workers=NUM_WORKERS,
                                                      pin_memory=True,
                                                      worker_init_fn=worker_init_fn,
                                                      collate_fn=CRFDataset.pad) if test_dataset is not None else None

        best_model = None
        best_model_transitions = None
        best_micro_f = 0
        best_model_ep = 0

        training_mode = []

        n_epoch = option.num_epoch
        if option.num_iteration is not None:
            iter_per_epoch = len(train_dataloader)
            n_epoch = max(option.num_iteration-1, 0)//iter_per_epoch+1

        do_valid_epoch = [n_epoch*(ep+1)//100 for ep in range(100)]
            
        if option.do_onlydict:
            training_mode.append(('onlydict', n_epoch))
        if option.do_pre:
            training_mode.append(('pre', n_epoch//10))
        training_mode.append(('main', n_epoch))
        if option.do_crfonly:
            training_mode.append(('crf-only', n_epoch))
        
        for train_mode, n_epoch in training_mode:
            lr = option.lr

            self.encoder.update_parameters(train_mode in ['main', 'crf-only'])
            only_dict = True if train_mode == 'onlydict' else False
        
            # optimizer and scheduler
            # https://tma15.github.io/blog/2020/10/03/%E8%87%AA%E7%84%B6%E8%A8%80%E8%AA%9E%E5%87%A6%E7%90%86-%E3%81%82%E3%81%AA%E3%81%9F%E3%81%AEbert%E3%81%AB%E5%AF%BE%E3%81%99%E3%82%8Bfine-tuning%E3%81%AF%E3%81%AA%E3%81%9C%E5%A4%B1%E6%95%97%E3%81%99%E3%82%8B%E3%81%AE%E3%81%8B-%E8%AB%96%E6%96%87%E7%B4%B9%E4%BB%8B/
            optimizer = torch.optim.AdamW(self.parameters(), lr=lr)
            if option.use_scheduler:
                lr_top = max(n_epoch//10, 1)
                scheduler = torch.optim.lr_scheduler.LambdaLR(
                    optimizer,
                    lr_lambda = lambda epoch: (epoch+1) / lr_top if epoch<lr_top else (n_epoch-epoch) / max(n_epoch-lr_top, 1),
                )

            # training loop
            start_time = time.time()
            for epoch in range(n_epoch):
                self.train()
                optimizer.zero_grad()
                tr_loss = 0.
                tr_loss_crf = 0.
                tr_loss_pointwise = 0.

                # balance for crf-loss and pointwise-loss
                lambda_pointwise = option.lambda_pointwise
                if train_mode in ['crf-only']:
                    lambda_pointwise = 0.

                lambda_crf = 1.0 - lambda_pointwise
                for step, batch in enumerate(train_dataloader):

                    batch = tuple(to_device(t, self.device, non_blocking=TO_NONBLOCKING) for t in batch)
                    input_ids, input_mask, segment_ids, predict_mask, label_ids, in_dict, full_label, not_full_pad = batch

                    loss = 0.

                    if lambda_crf > 0:
                        neg_log_likelihood = self.neg_log_likelihood(input_ids, segment_ids, input_mask, label_ids, in_dict, full_label, only_dict=only_dict)
                        loss += lambda_crf * neg_log_likelihood
                        if hasattr(neg_log_likelihood, 'item'):
                            tr_loss_crf += neg_log_likelihood.item() 

                    if lambda_pointwise > 0:
                        neg_log_likelihood_pointwise = self.neg_log_likelihood_pointwise(input_ids, segment_ids, input_mask, label_ids, in_dict, not_full_pad, only_dict=only_dict)
                        loss += lambda_pointwise * neg_log_likelihood_pointwise
                        if hasattr(neg_log_likelihood_pointwise, 'item'):
                            tr_loss_pointwise += neg_log_likelihood_pointwise.item()

                    if loss == 0.:
                        continue

                    loss.backward()
                    if hasattr(loss, 'item'):
                        tr_loss += loss.item()
                    optimizer.step()

                # End of epoch
                if option.use_scheduler:
                    scheduler.step()
                if not (epoch+1) in do_valid_epoch:
                    continue
                if not option.quiet_train:
                    print('({}) Epoch {}/{} Loss = {:.3f} (crf = {:.3f} / pointwise = {:.3f})'.format(train_mode, epoch+1, n_epoch, tr_loss / len(train_dataloader), tr_loss_crf / len(train_dataloader), tr_loss_pointwise / len(train_dataloader)))
                    elapsed_time = time.time()-start_time
                    print('Elapsed: {:.2f}s / Estimated: {:.2f}s'.format(elapsed_time, elapsed_time*(n_epoch-epoch-1)/(epoch+1)))
                    sys.stdout.flush()
                if valid_dataloader is not None:
                    if not option.quiet_valid:
                        print('Validation:')

                    eval_result = evaluate(self, valid_dataloader, quiet=option.quiet_valid)
                    micro_f = sum([r[0] for r in eval_result])/len(eval_result)
                    if micro_f > best_micro_f:
                        best_micro_f = micro_f
                        best_model = self.state_dict()
                        best_model_transitions = [self.transitions[i].data.clone() for i in range(len(self.transitions))]
                        best_model_ep = epoch+1
                    sys.stdout.flush()
            if best_model is not None:
                self.load_state_dict(best_model)
                for i in range(len(self.transitions)):
                    self.transitions[i].data = best_model_transitions[i].clone().data
                print('selected: {}'.format(best_model_ep))
        if test_dataloader is not None:
            if not option.quiet_test:
                print('Test:')
            evaluate(self, test_dataloader, quiet=option.quiet_test)

    def evaluate_file(self, datasets, batch_size):

        test_dataset = CRFDataset(datasets['test'], self.tokenizer, self.label_map, self.tag_index, dictionary=self.dictionary, max_seq_length=None) if 'test' in datasets and len(datasets['test']) > 0 else None
        test_dataloader = torch.utils.data.DataLoader(dataset=test_dataset,
                                                      batch_size=batch_size,
                                                      shuffle=False,
                                                      num_workers=NUM_WORKERS,
                                                      pin_memory=True,
                                                      worker_init_fn=worker_init_fn,
                                                      collate_fn=CRFDataset.pad) if test_dataset is not None else None
        evaluate(self, test_dataloader)

def evaluate(crf_model, predict_dataloader, quiet=False):

    crf_model.eval()
    with torch.no_grad():
        preds = []
        golds = []
        for i, batch in enumerate(predict_dataloader):
            batch = tuple(to_device(t, crf_model.device, non_blocking=TO_NONBLOCKING) for t in batch)
            input_ids, input_mask, segment_ids, predict_mask, label_ids, in_dict, _, not_full_pad = batch

            ret_list = crf_model(input_ids, segment_ids, input_mask, in_dict)

            predict_label = [
                [
                    [crf_model.label_list[j][li.item()] for li, mask_i in zip(predicted, mask) if mask_i]
                    for predicted, mask in zip(ret[1], predict_mask)
                ] for j, ret in enumerate(ret_list)
            ]
            gold_label = [
                [
                    [crf_model.label_list[j][li.item()] if li.item() >= 0 else '[PAD]' for li, mask_i in zip(gold, mask) if mask_i]
                    for gold, mask in zip(label_ids[j], predict_mask)
                ]
                for j in range(len(label_ids))
            ]

            for j in range(len(predict_label)):
                while len(preds) <= j:
                    preds.append([])
                    golds.append([])
                if gold_label[j] is not None:
                    preds[j].extend([predict_label[j][k] for k in range(len(predict_label[j])) if not_full_pad[k][j]])
                    golds[j].extend([gold_label[j][k] for k in range(len(predict_label[j])) if not_full_pad[k][j]])
        return [F1score(golds[j], preds[j], print_fullscore=not quiet, print_tagscore=not quiet) for j in range(len(golds))]

def define_crf_model(option, label_map, dictionary={}, device=None):

    if device is None:
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
    # tokenizer
    if hasattr(option, 'use_whitespace_split') and option.use_whitespace_split:
        tokenizer_file = option.whitespace_json_path
        tokenizer = PreTrainedTokenizerFast(
            tokenizer_file=tokenizer_file,
            bos_token="[CLS]",
            eos_token="[SEP]",
            unk_token="[UNK]",
            sep_token="[SEP]",
            pad_token="[PAD]",
            cls_token= "[CLS]",
            mask_token="[MASK]",
        )
    else:
        if option.bert_config is not None:
            tokenizer = AutoTokenizer.from_pretrained('/'.join(option.bert_config.split('/')[:-1]))
        else:
            tokenizer = AutoTokenizer.from_pretrained(option.bert_dir)

    # model definition
    if option.encoder_type == 'lstm':
        encoder = LSTMEncoder(vocab_size=tokenizer.vocab_size, padding_idx=tokenizer.pad_token_id)
    elif option.encoder_type == 'bert':
        if option.bert_config is not None:
            encoder = BERTEncoder(option.bert_config, by_config=True)
        else:
            encoder = BERTEncoder(option.bert_dir)
    else:
        raise ValueError('Unknown encoder type: {}'.format(option.encoder_type))
        exit()

    crf_model = CRF(encoder,
                    start_label_id=[label_map[i]['[CLS]'] for i in range(len(label_map))],
                    stop_label_id=[label_map[i]['[SEP]'] for i in range(len(label_map))],
                    label_map=label_map,
                    tokenizer=tokenizer,
                    use_dict=option.use_dict,
                    dictionary=dictionary,
                    device=device,
    )
    crf_model.option = option

    return crf_model

def save_crf_model(model, dirname, delete_file_path=True):

    os.makedirs(dirname, exist_ok=True)
    whitespace_json_path = model.option.whitespace_json_path
    with open('{}/label_map.txt'.format(dirname), 'w', encoding='utf-8') as f:
        json.dump(model.label_map, f)
    with open('{}/option.pickle'.format(dirname), 'wb') as f:
        if delete_file_path:
            model.option.delete_file_path()
        pickle.dump(model.option, f)
    with open('{}/ne.dict'.format(dirname), 'w', encoding='utf-8') as f:
        json.dump(model.dictionary, f)
    with open('{}/tag_index.txt'.format(dirname), 'w', encoding='utf-8') as f:
        json.dump(model.tag_index, f)
    if model.option.use_whitespace_split:
        os.makedirs('{}/tokenizer'.format(dirname), exist_ok=True)
        shutil.copyfile(whitespace_json_path, '{}/tokenizer/tokenizer.json'.format(dirname))
    else:
        os.makedirs('{}/tokenizer'.format(dirname), exist_ok=True)
        model.tokenizer.save_pretrained('{}/tokenizer/'.format(dirname))
        model.encoder.config.save_pretrained('{}/tokenizer/'.format(dirname))

    torch.save(model.state_dict(), dirname+'/crf.mdl')
    torch.save(model.transitions, dirname+'/crf_transitions.pt')
    
def load_crf_model(dirname, opts={}, device=None):

    if device is None:
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

    with open(dirname+'/label_map.txt', 'r', encoding='utf-8') as f:
        label_map = json.load(f)
    with open(dirname+'/option.pickle', 'rb') as f:
        option = pickle.load(f)
    with open(dirname+'/ne.dict', 'r', encoding='utf-8') as f:
        ne_dict = json.load(f)
    with open(dirname+'/tag_index.txt', 'r', encoding='utf-8') as f:
        tag_index = json.load(f)

    if hasattr(option, 'use_whitespace_split') and option.use_whitespace_split:
        local_filepath = '{}/tokenizer/tokenizer.json'.format(dirname)
        if os.path.isfile(local_filepath):
            opts['whitespace_json_path'] = local_filepath
        else:
            opts['whitespace_json_path'] = './tokenizer.json'
    else:
        local_filepath = '{}/tokenizer/config.json'.format(dirname)
        if os.path.isfile(local_filepath):
            opts['bert_config'] = local_filepath
        else:
            opts['bert_config'] = dirname
    option.update(opts)

    crf_model = define_crf_model(option, label_map, dictionary=ne_dict, device=device)
    crf_model.tag_index = tag_index
    crf_model.load_state_dict(torch.load(dirname+'/crf.mdl', map_location=torch.device(device)))
    crf_model.transitions = torch.load(dirname+'/crf_transitions.pt', map_location=torch.device(device))

    return crf_model

def worker_init_fn(worker_id):
    random.seed(worker_id)

class BaseEncoder(nn.Module):

    def __init__(self):

        super(BaseEncoder, self).__init__()

    def _get_features(self, input_ids, segment_ids, input_mask):

        raise NotImplementedError()

    def get_hidden_size(self):

        raise NotImplementedError()

    def _get_max_length(self):

        return None

    def update_parameters(self, flg):

        raise NotImplementedError()

    def save(self, filename):

        torch.save(self.state_dict(), filename)

    def load(self, filename):

        self.load_state_dict(torch.load(filename))

class BERTEncoder(BaseEncoder):

    def __init__(self, bert_dir, by_config=False):

        super(BERTEncoder, self).__init__()
        if by_config:
            self.config = AutoConfig.from_pretrained(bert_dir)
            self.bert = AutoModel.from_config(self.config)
        else:
            self.bert = AutoModel.from_pretrained(bert_dir)
            self.config = AutoConfig.from_pretrained(bert_dir)

    def _get_features(self, input_ids, segment_ids, input_mask):

        return self.bert(input_ids, token_type_ids=segment_ids, attention_mask=input_mask).last_hidden_state

    def get_hidden_size(self):

        return self.config.hidden_size

    def _get_max_length(self):

        return self.config.max_position_embeddings

    def update_parameters(self, flg):

        for pname, param in self.bert.named_parameters():
            param.requires_grad = flg

class LSTMEncoder(BaseEncoder):

    def __init__(self, vocab_size, d_emb = 128, padding_idx = -100, dropout_rate = 0.25):

        super(LSTMEncoder, self).__init__()
        self.embedding = nn.Embedding(num_embeddings = vocab_size,
                                      embedding_dim = d_emb,
                                      padding_idx = padding_idx,
        )
        self.LSTM = nn.LSTM(input_size = d_emb,
                            hidden_size = d_emb//2,
                            num_layers = 1,
                            bias = True,
                            batch_first = False,
                            dropout = dropout_rate,
                            bidirectional = True,
        )

    def _get_features(self, input_ids, segment_ids, input_mask):

        input_len = [sum(mask_i).data for mask_i in input_mask]
        embed = self.embedding(input_ids)
        packed_embed = pack_padded_sequence(embed, input_len, enforce_sorted=False, batch_first=True)
        packed_output, (hn, cn) = self.LSTM(packed_embed)
        output, seq_len = pad_packed_sequence(packed_output, batch_first=True)
        return output

    def get_hidden_size(self):

        return self.LSTM.hidden_size * (2 if self.LSTM.bidirectional else 1)

    def update_parameters(self, flg):

        for layer in [self.embedding, self.LSTM]:
            for param in layer.parameters():
                param.requires_grad = flg

if __name__=='__main__':

    pass
