#-*- coding: utf-8 -*-

import sys

import torch

from crf import define_crf_model, load_crf_model, save_crf_model
from crf import ModelOption
from crf import make_whitespace_tokenizer
from corpus_reader import load_dataset

def train_ner(filenames, option):

    datasets, label_set, ne_dict = load_dataset(filenames)

    '''
    for i, d in enumerate(ne_dict):
        for k, v in d.items():
            for vi in v:
                datasets['train'].append((vi, [[k+'-B' if j==0 else k+'-I' for j in range(len(vi))] if i==ti else ['[PAD]' for _ in vi] for ti in range(len(ne_dict))]))
    '''

    if option.target_tag is not None and len(option.target_tag) > 0:
        target_tag = option.target_tag
        for data_type, data_list in datasets.items():
            dataset_target = []
            for words, tags in data_list:
                tags_target = [t for i, t in enumerate(tags) if i in target_tag]
                dataset_target.append((words, tags_target))
            datasets[data_type] = dataset_target

        label_set = [l for i, l in enumerate(label_set) if i in target_tag]
        ne_dict = [d for i, d in enumerate(ne_dict) if i in target_tag]

    label_map = [{'O': 0, '[CLS]': 1, '[SEP]': 2, 'X': 3, '[PAD]': -100} for _ in label_set]
    for i in range(len(label_map)):
        cur_label_id = 4
        for label in label_set[i]:
            if not label in label_map[i]:
                label_map[i][label] = cur_label_id
                cur_label_id += 1
    sys.stdout.flush()

    if option.use_whitespace_split:
        corpus = [elem[0] for elem in datasets['train']]
        tmp_tokenizer_json = '/tmp/CaPSL/tokenizer.json'
        make_whitespace_tokenizer(corpus, vocab_size=option.vocab_size, do_strip_accents=option.do_strip_accents, save_filename=tmp_tokenizer_json)
        opts = {'whitespace_json_path': tmp_tokenizer_json}
        option.update(opts)

    crf_model = define_crf_model(option, label_map, dictionary=ne_dict)
    crf_model.train_ner(
        datasets,
        label_set,
        ne_dict,
        option,
    )
    return crf_model

def init_random_seed(seed=42, deterministic=False):
        
    # initialization
    import torch
    import random
    import numpy as np
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = deterministic

if __name__=='__main__':

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--seed', type=int, default=42)
    parser.add_argument('--model_dir')

    # mode
    parser.add_argument('--train', action='store_true')
    parser.add_argument('--inference', action='store_true')
    parser.add_argument('--evaluate', action='store_true')

    # training hyper-parameter
    parser.add_argument('--num_epoch', type=int, default=20)
    parser.add_argument('--num_iteration', type=int, default=None)
    parser.add_argument('--lr', type=float, default=2e-6)
    parser.add_argument('--encoder', choices=['bert', 'lstm'], default='bert')
    parser.add_argument('--use_dict', action='store_true')
    parser.add_argument('--batch_size', type=int, default=32)
    parser.add_argument('--use_scheduler', action='store_true')
    parser.add_argument('--bert_dir', default='bert-base-uncased')
    parser.add_argument('--lambda_pointwise', type=float, default=0.5)
    parser.add_argument('--use_whitespace_split', action='store_true')
    parser.add_argument('--vocab_size', type=int, default=32_000)
    parser.add_argument('--do_strip_accents', action='store_true')

    # evaluation hyper-parameter
    parser.add_argument('--eval_batch_size', type=int, default=64)

    # training mode
    parser.add_argument('--do_pre', action='store_true')
    parser.add_argument('--do_onlydict', action='store_true')
    parser.add_argument('--do_crfonly', action='store_true')
    parser.add_argument('--target_tag', type=int, nargs='*', default=None)

    # data files
    parser.add_argument('--file_format', choices=['spl', 'conll'], default='spl')
    ## for training
    parser.add_argument('--trainfile', action='append', nargs='*', default=[])
    parser.add_argument('--dictfile', action='append', nargs='*', default=[])
    parser.add_argument('--validfile', action='append', nargs='*', default=[])
    parser.add_argument('--testfile', action='append', nargs='*', default=[])
    ## for inference
    parser.add_argument('--target', default=[])

    # output verbose on training loop
    parser.add_argument('--quiet_train', action='store_true')
    parser.add_argument('--quiet_valid', action='store_true')
    parser.add_argument('--quiet_test', action='store_true')

    args = parser.parse_args()

    # seed
    init_random_seed(
        seed=args.seed,
        deterministic=True
    )

    if args.train:

        filenames = {
            'format': args.file_format,
            'train': sum(args.trainfile, []),
            'dict': sum(args.dictfile, []),
            'valid': sum(args.validfile, []),
            'test': sum(args.testfile, []),
        }
        model_option = {
            'bert_dir': args.bert_dir,
            'encoder_type': args.encoder,
            'batch_size':args.batch_size,
            'eval_batch_size': args.eval_batch_size,
            'lr':args.lr,
            'use_scheduler':args.use_scheduler,
            'num_epoch':args.num_epoch,
            'num_iteration':args.num_iteration,
            'lambda_pointwise':args.lambda_pointwise,
            'quiet_train':args.quiet_train,
            'quiet_valid':args.quiet_valid,
            'quiet_test':args.quiet_test,
            'do_pre':args.do_pre,
            'do_onlydict':args.do_onlydict,
            'do_crfonly':args.do_crfonly,
            'target_tag':args.target_tag,
            'use_dict':args.use_dict,
            'use_whitespace_split':args.use_whitespace_split,
            'vocab_size':args.vocab_size,
        }
        option = ModelOption(model_option)

        model = train_ner(filenames, option)
        if args.model_dir:
            save_crf_model(model, args.model_dir)

    if args.inference:

        if args.model_dir is None:
            sys.stderr.write('Error: specify model directory\n')
            exit(1)
        if args.target is None:
            sys.stderr.write('Error: specify target file\n')
            exit(1)

        model = load_crf_model(args.model_dir)
        model.eval()

        with open(args.target, 'r') as f:
            texts = [l.strip() for l in f]
            labels = model.predict(texts, batch_size=args.batch_size)

            for text, label in zip(texts, labels):

                word_label = ['/'.join([w] + [label[i][j] for i in range(len(label))]) for j, w in enumerate(text.split(' '))]
                print(' '.join(word_label))

    if args.evaluate:

        if args.model_dir is None:
            sys.stderr.write('Error: specify model directory\n')
            exit(1)
        if args.testfile is None:
            sys.stderr.write('Error: specify test file\n')
            exit(1)

        filenames = {
            'format': args.file_format,
            'train': sum(args.trainfile, []),
            'dict': sum(args.dictfile, []),
            'valid': sum(args.validfile, []),
            'test': sum(args.testfile, []),
        }
        datasets, label_set, ne_dict = load_dataset(filenames)

        model = load_crf_model(args.model_dir)
        model.evaluate_file(datasets, args.eval_batch_size)
